package ims.coref.resolver;

import java.util.HashSet;
import java.util.List;
import java.util.Set;

import de.bwaldvogel.liblinear.FeatureNode;

import ims.coref.Coref;
import ims.coref.data.CorefSolution;
import ims.coref.data.Document;
import ims.coref.data.PairInstance;
import ims.coref.data.Span;
import ims.coref.data.SpanListStruct;
import ims.coref.features.FeatureSet;
import ims.coref.markables.IMarkableExtractor;
import ims.coref.postprocessor.IChainPostProcessor;
import ims.ml.liblinear.LibLinearModel;

public abstract class AbstractResolver implements ICoreferenceResolver{

	public static enum SingleLinkConstraint { None, NoTransitiveEmbedding, FirstOnly }
	
	protected final IMarkableExtractor me;
	protected final LibLinearModel llModel;
	protected final FeatureSet fs;
	public final IChainPostProcessor pp;
		
//	public final EvaluateMarkables evalMarkables=new EvaluateMarkables();
//	private final GoldStandardChainExtractor gsce=new GoldStandardChainExtractor();
	protected final SingleLinkConstraint slc;
	
	final int decodeWindow;
	
	int chainIdCounter=0;
	
	protected AbstractResolver(IMarkableExtractor me,LibLinearModel llModel,FeatureSet fs,IChainPostProcessor pp,int decodeWindow,SingleLinkConstraint slc){
		this.me=me;
		this.llModel=llModel;
		this.fs=fs;
		this.pp=pp;
		this.decodeWindow=decodeWindow;
		this.slc=slc;
	}

	
	@Override
	public CorefSolution resolve(Document d) {
		chainIdCounter=0;
		Set<Span> predSpans=me.extractMarkables(d);
//
//		List<Span> spans=new ArrayList<Span>(predSpans);
//		Collections.sort(spans);
		SpanListStruct sls=SpanListStruct.fromCollection(predSpans);
		CorefSolution solution = doResolve(sls,d);
		pp.postProcess(solution, d, sls);
		return solution;
	}
	
	void resolveBestLink(SpanListStruct sls,CorefSolution cs,int anaIndex,double th,Document d){
		switch(slc){
		case None:  				resolveBestLinkNoConstraint(sls,cs,anaIndex,th,d); return;
		case NoTransitiveEmbedding: resolveBestLinkNoTransitiveEmbedding(sls,cs,anaIndex,th,d); return;
		case FirstOnly:  			resolveBestLinkFirstOnlyConstraint(sls,cs,anaIndex,th,d); return;
		default: throw new Error("not implemented");
		}
	}
	void resolveClosestFirst(SpanListStruct sls,CorefSolution cs,int anaIndex,Document d) {
		switch(slc){
		case None: 					resolveClosestFirstNoConstraint(sls,cs,anaIndex,d); return;
		case NoTransitiveEmbedding: resolveClosestFirstNoTransitiveEmbedding(sls,cs,anaIndex,d); return;
		case FirstOnly: 			resolveClosestFirstFirstOnlyConstraint(sls,cs,anaIndex,d); return;
		default: throw new Error("not implemented");
		}
	}
	void resolveClosestFirstNoConstraint(SpanListStruct sls,CorefSolution cs,int anaIndex,Document d){
		Span anaphor=sls.get(anaIndex);
		for(int antIndex=anaIndex-1,dist=1;antIndex>=0 && dist<decodeWindow;--antIndex,dist++){
			Span ant=sls.get(antIndex);
			PairInstance pi=new PairInstance(ant,anaphor,sls.getMentionDist(antIndex, anaIndex),sls.getNesBetween(antIndex, anaIndex));
			List<FeatureNode> fns=fs.getFeatureNodes(pi,d);
			int cl=llModel.getMostProbableClass(fns);
			if(cl==Coref.POSITIVE){
				cs.addLink(ant,anaphor);
				return;
			}
		}
	}
	void resolveClosestFirstFirstOnlyConstraint(SpanListStruct sls,CorefSolution cs,int anaIndex,Document d){
		Set<Integer> skip=new HashSet<Integer>();
		Span anaphor=sls.get(anaIndex);
		for(int antIndex=anaIndex-1,dist=1;antIndex>=0 && dist<decodeWindow;--antIndex,dist++){
			Span ant=sls.get(antIndex);
//			Integer antChainID=cs.span2int.get(ant);
			Integer antChainID=cs.getSpanChainID(ant);
			if(antChainID!=null)
				if(skip.contains(antChainID))
					continue;
				else
					skip.add(antChainID);
			PairInstance pi=new PairInstance(ant,anaphor,sls.getMentionDist(antIndex, anaIndex),sls.getNesBetween(antIndex, anaIndex));
			List<FeatureNode> fns=fs.getFeatureNodes(pi,d);
			int cl=llModel.getMostProbableClass(fns);
			if(cl==Coref.POSITIVE){
				cs.addLink(ant,anaphor);
				return;
			}
		}
	}
	void resolveClosestFirstNoTransitiveEmbedding(SpanListStruct sls,CorefSolution cs,int anaIndex,Document d){
		Set<Integer> skip=new HashSet<Integer>(); //store chain ids of spans that embed this anaphor here, so that if we visit other spans from that chain later on, we ignore it.
		Span anaphor=sls.get(anaIndex);
		for(int antIndex=anaIndex-1,dist=1;antIndex>=0 && dist<decodeWindow;--antIndex,dist++){
			Span ant=sls.get(antIndex);
			{ //This code skips
				Integer antChainID=cs.getSpanChainID(ant);
				if(ant.embeds(anaphor)){
//					Integer antChainID=cs.span2int.get(ant);
					
					if(antChainID!=null)
						skip.add(antChainID);
				} else {
					if(!skip.isEmpty()){
//						Integer antChainID=cs.span2int.get(ant);
						if(antChainID!=null && skip.contains(antChainID))
							continue;
					}
				}
			}
			PairInstance pi=new PairInstance(ant,anaphor,sls.getMentionDist(antIndex, anaIndex),sls.getNesBetween(antIndex, anaIndex));
			List<FeatureNode> fns=fs.getFeatureNodes(pi,d);
			int cl=llModel.getMostProbableClass(fns);
			if(cl==Coref.POSITIVE){
				cs.addLink(ant,anaphor);
				return;
			}
		}
	}
	void resolveBestLinkFirstOnlyConstraint(SpanListStruct sls,CorefSolution cs,int anaIndex,double th,Document d){
		Set<Integer> skip=new HashSet<Integer>();
		Span anaphor=sls.get(anaIndex);
		Span bestSpan=null;
		double bestScore=-1;
		for(int antIndex=anaIndex-1,dist=1;antIndex>=0 && dist<decodeWindow;--antIndex,++dist){
			Span ant=sls.get(antIndex);
//			Integer antChainID=cs.span2int.get(ant);
			Integer antChainID=cs.getSpanChainID(ant);
			if(antChainID!=null){
				if(skip.contains(antChainID))
					continue;
				else
					skip.add(antChainID);
			}
			PairInstance pi=new PairInstance(ant,anaphor,sls.getMentionDist(antIndex, anaIndex),sls.getNesBetween(antIndex, anaIndex));
			List<FeatureNode> fns=fs.getFeatureNodes(pi,d);
			double p=llModel.getProbabilityForClass(fns,Coref.POSITIVE);
			if(p>=th && p>bestScore){
				bestScore=p;
				bestSpan=ant;
			}
		}
		if(bestSpan!=null){
			cs.addLink(bestSpan,anaphor);
		}

	}
	void resolveBestLinkNoTransitiveEmbedding(SpanListStruct sls,CorefSolution cs,int anaIndex,double th,Document d){
		Set<Integer> skip=new HashSet<Integer>();
		Span anaphor=sls.get(anaIndex);
		Span bestSpan=null;
		double bestScore=-1;
		for(int antIndex=anaIndex-1,dist=1;antIndex>=0 && dist<decodeWindow;--antIndex,++dist){
			Span ant=sls.get(antIndex);
			{ //This code skips -- copy and paste from above
				Integer antChainID=cs.getSpanChainID(ant);
				if(ant.embeds(anaphor)){
					if(antChainID!=null)
						skip.add(antChainID);
				} else {
					if(!skip.isEmpty()){
						if(antChainID!=null && skip.contains(antChainID))
							continue;
					}
				}
			}
			PairInstance pi=new PairInstance(ant,anaphor,sls.getMentionDist(antIndex, anaIndex),sls.getNesBetween(antIndex, anaIndex));
			List<FeatureNode> fns=fs.getFeatureNodes(pi,d);
			double p=llModel.getProbabilityForClass(fns,Coref.POSITIVE);
			if(p>=th && p>bestScore){
				bestScore=p;
				bestSpan=ant;
			}
		}
		if(bestSpan!=null){
			cs.addLink(bestSpan,anaphor);
		}
	}
	void resolveBestLinkNoConstraint(SpanListStruct sls,CorefSolution cs,int anaIndex,double th,Document d){
		Span anaphor=sls.get(anaIndex);
		Span bestSpan=null;
		double bestScore=-1;
		for(int antIndex=anaIndex-1,dist=1;antIndex>=0 && dist<decodeWindow;--antIndex,++dist){
			Span ant=sls.get(antIndex);
			PairInstance pi=new PairInstance(ant,anaphor,sls.getMentionDist(antIndex, anaIndex),sls.getNesBetween(antIndex, anaIndex));
			List<FeatureNode> fns=fs.getFeatureNodes(pi,d);
			double p=llModel.getProbabilityForClass(fns,Coref.POSITIVE);
			if(p>=th && p>bestScore){
				bestScore=p;
				bestSpan=ant;
			}
		}
		if(bestSpan!=null){
			cs.addLink(bestSpan,anaphor);
		}
	}
	

	public String toString(){
		return this.getClass().getCanonicalName();
	}
//	abstract CorefSolution doResolve(List<Span> spans,Document d);
	abstract CorefSolution doResolve(SpanListStruct sls,Document d);
	
	@Override
	public IChainPostProcessor getPostProcessor() {
		return pp;
	}
}
